
import numpy as np
import math

# ************************************************************************
def Offset_Correction(Value, Velocity_Offset):
#
# Description:    Correct a binary velocity value if an offset is attached
# -----------     to it.
#
#                 Input: Value = binary data value read from file
#                        Velocity_Offset, binary value read from the data file
#                        Nb_Gates, binary value read from the data file

    Value = Value + Velocity_Offset
    ICOR = 0
    if Value > 127:
      ICOR = -256
    if Value < -128:
      ICOR = 256
    Value = Value + ICOR
    Value = float(Value - Velocity_Offset)        
    
    return Value

# ************************************************************************
def Get_No_Channel(fid):
#
# Description:    Return the number of the channel attached to the data
# -----------     profile.
#                 if the returned value is = -1, the file limit was reached
#
#                 Input: the file pointer point the beginning of a profile
#                 Output: >0 the channel number
#                         -1 the channel was not found

    try:
      N = np.fromfile(fid, dtype=np.int16, count=1, sep="")[0]         # N = Nb nytes in profile
      m = N-5
      fid.seek(m, 1)         # read from current position in file
      No_Channel = np.fromfile(fid, dtype=np.int8, count=1, sep="")[0]         # N = Nb nytes in profile
      # Correct error in previous version....
      if No_Channel == 0:
        No_Channel = 1
      # Replace the file pointer at the beginning of the profile     
      fid.seek(-(N-2), 1)
    except:
      No_Channel = -1
    
    return No_Channel

# ************************************************************************
def Get_No_Block(fid):
#
# Description:    Return the number of the block attached to the data
# -----------     profile.
#
#                 Input: the file pointer point the beginning of a profile
#                 Output: >0 the block number
#                         -1 the channel was not found
#
    try:
      N = np.fromfile(fid, dtype=np.int16, count=1, sep="")[0]         # N = Nb nytes in profile
      fid.seek(N-10, 1)
      No_Block = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]         # N = Nb nytes in profile
      # Replace the file pointer at the begining of the profile  
      fid.seek(-(N-6), 1)
    except:
      No_Block = -1  

    return No_Block

# ************************************************************************
def Get_Number_Of_Mpx_Used_Channels(Mpx_Mode):
#
# Description:    Return the number oo used channel
# -----------     
#
#                 Input: Mpx_Mode
#                 Output: Number of used channels in Mpx mode actif
#
    Nb_Used = 0
    imax = 0x4000
    i = 0x8
    while i != imax:
      i = i << 1
      if (Mpx_Mode & i) != 0:
        Nb_Used = Nb_Used + 1

    return Nb_Used

# ************************************************************************
def Get_Dop_Profiles(fid, Graph, Channel, Block):
#
# Version: 3.2
# ------------
# Take into acount flow rate, auto correction of the aliasing and new format
# from UDOP version 4.00.1
#
#
# Description:    This function extracts from an UDOP binary data file all
# -----------     the data profiles contained in single graph for a specific
#                 channel and a specific block.
#
#                 The function checks if the handle associated to the data
#                 file points to a correct UDOP binary data file. The
#                 output variable Np contains the error code.
#
#                 UDOP version supported: From 2.00.1
#
# Input:          fid   : a handle to an already opened binary data file
#                 Graph : number of the graph (1,2,3,..)
#                         each profile may contain more than one graph, for
#                         instance, the velocity profile and the echo
#                         profile.
#                 Channel: Number of the channel
#                          for DOP3000 must be set to 1
#                          In case of UDV 2D or 3D not relevant (set to 10)
#                 Block  : number of data block
#
#
# Output:         X  : a 1 dimension data table that gives the values of
#                      the X abscissa for the selected graph. These
#                      values are depths in mm, but can be velocities in
#                      mm/s in case of the power spectrum of a single gate
#
#                 Y  : a 2 dimensions data table that contains the values of
#                      the measured data. Each row is associated to a profile
#                      and each column is associated to a gate.
#
#                      In case of raw data acquisition only:
#                      row:           data
#                       1     first emission, I signal
#                       2     first emission, Q signal
#                       3     second emission, I signal
#                       4     second emission, Q signal
#                                     etc....
#
#                 Np : if > 0  number of available data profiles
#                              (= number of rows of the Y table,
#                               or number of emissions in case of
#                               raw data acquisition)
#                      if < 0  error code
#                          -1 : not an UDOP binary data file
#                          -2 : unsupported software version
#                          -3 : the selected graphs does not exists
#                          -4 : the selected channel does not exists
#                          -5 : the selected block does not exists
#
#              Gates : Number of gates contained in the profile
#
#               Info : a 2 dimension data table that gives the additionnal
#                      information associated to each profile. The number
#                      of rows equal the number of rows of the Y table
#
#                       Column 1 : the time stamp in ms
#                                  (in micro seonds in case of raw data
#                                  acquisition)
#                       Column 2 : number of the block
#                       Column 3 : no meaning (available for future developement)
#                       Column 4 : no meaning (available for fnnuture developement)
#                       Column 5 : no meaning (available for future developement)
#                       Column 6 : number of the channel
#
# Structure of a binary profile
# -----------------------------
# A : 1 word that gives the total number of bytes contained in the profile
# B : 1 word that gives the number of bytes contained in the follwing graph
#     if 0 no more graph in the profile
# C : 1 byte that identify the type of data type contained in the graph
#     see table "Data Type constant" line 115 in this file
# D : the data bytes of the graph
#     if each data are in byte format B = number of bytes contained in the
#     graph
# Struture B,C,D repeat until B = 0
# E : 1 word = 0 indicates that no more exists
# F : 1 doubleword = time stamps in ms*10
# G : 1 word = number of the block
# H : 1 byte reserved for further use
# I : 1 byte reserved for further use
# J : 1 byte reserved for further use
# K : 1 word that gives the total number of bytes contained in the profile
#     (K = A)
#
# Date: 21 July 2016
# Author : J.-Cl Willemetz, Signal Processing SA
#          For any questions or comments please email to:
#          contact@signal-processing.com
# -------------------------------------------------------------------------
    fid.seek(0)
#
# Check if the file comes from UDOP software
#
    File_Ident = fid.read(7)
    S=File_Ident.decode('utf-8')
    if S !='BINUDOP':
      return 1,1,-1,0,0
#
# Get the software version of file
#
    S = fid.read(1)
    V = fid.read(6)
    V =V.decode('utf-8')
    V1=int(V[0])
    V2=int(V[2:4])
    V3=int(V[5])
    Version = 1000*V1 + 10*V2 + V3;    
    if Version < 2001:
      return 1,1,-2,0,0
    
# Some UDOP constants -----------------------------------
#
#      Data Type constant
#      -------------------
#      These constants define the type of data contains in a graph
#      The data format is 8 bit except for the following data type values:
#      4, 16, 17, 19, 24, 25, 30 : the data in format 16 bits
#      18 : data in format 32 bits
#
    pi = 3.1416
    
    Data_Type_Velocity = 0
    Data_Type_Echo = 1
    Data_Type_Energy = 2
    Data_Type_Gate_FFT = 3
    Data_Type_Phase_deg_10 = 4           # phase in Deg*10, format 16 bits
    Data_Type_Vitson = 5
    Data_Type_Frequency = 6
    Data_Type_Frequency_TR1 = 7
    Data_Type_Frequency_TR2 = 8
    Data_Type_Frequency_TR3 = 9
    Data_Type_Echo_TR1 = 10
    Data_Type_Echo_TR2 = 11
    Data_Type_Echo_TR3 = 12
    Data_Type_Energy_TR1 = 13
    Data_Type_Energy_TR2 = 14
    Data_Type_Energy_TR3 = 15
    Data_Type_Velocity_mm_s_10 = 16       # Velocity in mm/s * 10
    Data_Type_Velocity_mm_s = 17          # Velocity in mm/s
    Data_Type_Flow_ml_min = 18            # Flowrate in ml/min,  format 32 bits
    Data_Type_Diameter_10 = 19            # diameter in mm*10, format 16 bits
    Data_Type_Aliasing_Refer = 20         # reference profil base on 2 PRF
    Data_Type_Aliasing_Refer_TR1 = 21     # reference profil base on 2 PRF for TR1
    Data_Type_Aliasing_Refer_TR2 = 22     # reference profil base on 2 PRF for TR2
    Data_Type_Aliasing_Refer_TR3 = 23     # reference profil base on 2 PRF for TR3
    Data_Type_TGC = 28
    Data_Type_IQ = 29
    Data_Type_Frequency_Hz_10 = 24
    Data_Type_Depth_mm_10 = 25            # depths [mm*10]
    Data_Type_Angle_Deg_10 = 30           # angle [deg*10], format 16 bits
# ------------------------
    Nb_Para_Fct = 256                     # Number of bytes of the table of parameters associated to a channel
    FFT_Scale_Pmin = -50                   # FFT minimum level [dBm]
       
    Retard_Flt_50k = 10500                 # IQ filter delay [ns]
    Retard_Flt_100k = 6300
    Retard_Flt_150k = 3500
    Retard_Flt_200k = 2800
    Retard_Flt_250k = 2100
    Retard_Flt_300k = 1800

    Retard_Analog_DOP4000 = 1200
       
    Ofs_Data = 31268  # offset in file of the first data value

    Nb_Gates = 0
    
# -------------------------------------------------------------------------
#      Check if the channel is known in the file
#
    fid.seek(548, 0)          # point to the parameters table of channel 1
    Parameters = np.fromfile(fid, dtype=np.uint32, count=Nb_Para_Fct, sep="")

    # Determine the instrument (detection based on the frequency of the AD converter)
    Freq_AD = Parameters[29]
    if (Freq_AD % 256) != 3:
      Dop_Model = 3010
    else:
      Dop_Model = 4000

    Mpx_Mode = Parameters[52]

    if Mpx_Mode & 0x00000002 != 0:
      # if the multiplexer was active ....
      # verify that the selected channel is included in the mpx sequence
      if Dop_Model == 4000:
        if Channel > 4:
          return 1,1,-4
        
      jt = 1
      jt = jt << Channel+3
      if Mpx_Mode & jt == 0:
        return 1,1,-4
    
    else:
      #Verify that the selected channel exists (non multiplex mode)
      if Mpx_Mode & 0x00000004 != 0:
        # if in UDV 2D/3D mode, the channel number = 10
        Channel = 10
      else:
        fid.seek(Ofs_Data,0)
        Channel_read = Get_No_Channel(fid)
        if Channel_read != Channel:
          return 1,1,-4
        
#
#   Select the parameters for the selected channel
# 
    fid.seek(548 + 1024*(Channel-1), 0)
    Parameters = np.fromfile(fid, dtype=np.int32, count=Nb_Para_Fct, sep="")
#
    i = -1
    Freq_Emi = Parameters[i+1]     # Emitting frequency in kHZ
    Prf_Periode_us = Parameters[i+6] # PRF in micro seconds
    Indice_First_Gate = Parameters[i+10]
    Skip_Gate = Parameters[i+11] # time between gates indice
    Nb_Gates = Parameters[i+14] #
    Nprf = Parameters[i+15] #
    Scale_Angle = Parameters[i+16] # velocity scaling factor
    Sound_Speed = Parameters[i+20] # sound velocity in m/s
    Doppler_Angle = Parameters[i+21] # Doppler angle in degrees
    Module_Scale = Parameters[i+22]     # Emitting frequency in kHZ
    Velocity_Offset = Parameters[i+23] # Coded value of the velocity Offset
    No_Flt_ispPAC = Parameters[i+28]   # IQ filter indice
    Freq_AD_Info    = Parameters[i+30] # fixe the AD frequency
    FFT_Scale_Pmax = Parameters[i+31]
    # Mpx_Mode = np.uint32( np.int32(Parameters[i+53]) )
    Freq_Latch_Tgc_kHz = Parameters[i+41] / 1000.0
    UDOP_Fct_Mode =  Parameters[i+44]
    Cursor_On_Gate = Parameters[i+61]
       
    Tr1_Frequency_Offset = Parameters[i+71]
    Tr2_Frequency_Offset = Parameters[i+72]
    Tr3_Frequency_Offset = Parameters[i+73]       
    MD_Scale_Angle_Tr1 = Parameters[i+74]
    MD_Scale_Angle_Tr2 = Parameters[i+75]
    MD_Scale_Angle_Tr3 = Parameters[i+76]       
    MD_Module_Scale_Tr1 = Parameters[i+77]     # Emitting frequency in kHZ
    MD_Module_Scale_Tr2 = Parameters[i+78]     # Emitting frequency in kHZ
    MD_Module_Scale_Tr3 = Parameters[i+79]     # Emitting frequency in kHZ

    Acqui_IQ_Nprf = Parameters[i+91]
    Acqui_IQ_Nb_Gates = Parameters[i+92]
    Acqui_IQ_First_Gate = Parameters[i+93]
       
    Over_All_Gain_dB = Parameters[i+94]        # gain for FFT
    Flow_unit = Parameters[i+112]              # 0 : ml/min
                                                 # 1 : ml/s
                                                 # 2 : dl/s
    Flow_Scale = Parameters[i+113]       
    Aliasing_Ctrl_1 = Parameters[i+114]
    Aliasing_Ctrl_2 = Parameters[i+115]

    Sound_Speed_Wall = Parameters[i+116]
    Length_Wall = Parameters[i+117]
    Sound_Speed_Couplant = Parameters[i+118]
    Length_Couplant = Parameters[i+119]
    Div_Freq_AD = Parameters[i+121]

    
# -------------------------------------------------------------------------
#   Compute some usefull variables
#

    if Dop_Model == 4000:
      Retard_Flt_IQ_ns = 1200
      Freq_AD = 1000 * 400 /(2*Div_Freq_AD);
    else:
      if No_Flt_ispPAC == 0:
        Retard_Flt_IQ_ns = Retard_Flt_50k
      if No_Flt_ispPAC == 1:
        Retard_Flt_IQ_ns = Retard_Flt_100k
      if No_Flt_ispPAC == 2:
        Retard_Flt_IQ_ns = Retard_Flt_150k
      if No_Flt_ispPAC == 3:
        Retard_Flt_IQ_ns = Retard_Flt_200k
      if No_Flt_ispPAC == 4:
        Retard_Flt_IQ_ns = Retard_Flt_250k
      if No_Flt_ispPAC == 5:
        Retard_Flt_IQ_ns = Retard_Flt_300k

      if mod(Freq_AD_Info,256) == 0:
        Freq_AD = 6000
      if mod(Freq_AD_Info,256) == 1:
        Freq_AD = 12000
      if mod(Freq_AD_Info,256) == 2:
        Freq_AD = 40000

    Cos_AngDop = math.cos(Doppler_Angle*pi/180)
    Coded_Value_To_Frequency = 1000 * Scale_Angle/(256*pi*Prf_Periode_us)
    Coded_Value_To_Velocity = 128 * Coded_Value_To_Frequency * Sound_Speed /(256*Freq_Emi * Cos_AngDop)
    Coded_Value_To_Frequency_For_Tr1 = 1000 * MD_Scale_Angle_Tr1 /(256*pi*Prf_Periode_us)
    Coded_Value_To_Frequency_For_Tr2 = 1000 * MD_Scale_Angle_Tr2 /(256*pi*Prf_Periode_us)
    Coded_Value_To_Frequency_For_Tr3 = 1000 * MD_Scale_Angle_Tr3 /(256*pi*Prf_Periode_us)

    if (UDOP_Fct_Mode & 2) > 0:
      if (UDOP_Fct_Mode & 1) > 0:
        Coupling_Time_us = 1000 * ((Length_Couplant*0.1/Sound_Speed_Couplant) + (Length_Wall*0.1/Sound_Speed_Wall))
      else:
        Coupling_Time_us = 1000 * ((Length_Couplant*0.2/Sound_Speed_Couplant) + (Length_Wall*0.2/Sound_Speed_Wall))
    else:
      Coupling_Time_us = 0.0
      
    Number_Of_Mpx_Used_Channels = Get_Number_Of_Mpx_Used_Channels(Mpx_Mode)
      
# -------------------------------------------------------------------------
# Get or define the measurement depths
#
    fid.seek(Ofs_Data, 0)
    X = []
    if Version >= 3001:

#     Point to the selected channel
#     The first profiles contain the depths of the gates. These are pseudo profiles
#     as they do not contain measured data. It exists one pseudo profile for
#     each curves contained in a profile data line.
#     In case of acquisition in simultaneous multiplexing mode, each block contains a number of
#     pseudo profiles containing the depths values equal to number of measured channels. 
#
      Channel_read = Get_No_Channel(fid)
      if Channel_read < 0:
        return 1,1,-4
      while Channel_read != Channel:
        N = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]
        fid.seek(N-2, 1)
        Channel_read = Get_No_Channel(fid)
        if Channel_read < 0:
          return 1,1,-4

#     Point to the selected graph
#     Each profile can contain many graphs. The data contained in each graph starts by
#     1 word that gives the number of data bytes followed by one byte that
#     gives the type of measured data (see data type constant above)
      Nb_Bytes_In_Profile = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]
      Gr = 1
      while (Gr < Graph):
        Gr = Gr + 1
        Nb_Bytes_In_Graph = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]
        if Nb_Bytes_In_Graph == 0:
          return 1,1,-3
        fid.seek(Nb_Bytes_In_Graph + 1, 1)

      Nb_Bytes_In_Graph = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]
      if Nb_Bytes_In_Graph == 0:
        return 1,1,-3

      # The selected graph is found ....
      Data_Type = np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0]
      N = int(Nb_Bytes_In_Graph/2)              # assume 16 bit data format
      
      if Data_Type == Data_Type_Velocity_mm_s:     # in case of gate_FFT
        # get velocity values
        for i in range (0, N):
          Vabs = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]  
          X.append(Vabs)
      else:
        if Data_Type == Data_Type_Diameter_10:
          # get diameter for flow rate
          Vabs = np.fromfile(fid, dtype=np.int32, count=1, sep="")[0]  # for other cases it is depths values
          X.append(Vabs)
        else:
          for i in range (0, N):
            # get depths values and place in millimeters 
            Vabs = 0.1 * np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]  
            X.append(Vabs)
        
      # Determine the position in the file of the first value profile
      fid.seek(Ofs_Data, 0)
      Ofs_First_Value = Ofs_Data
      if (Mpx_Mode & 0x2) != 0:
        # en Mpx actif ....
        for i in range (0, Number_Of_Mpx_Used_Channels):
          N = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]
          Ofs_First_Value = Ofs_First_Value + N
          fid.seek(N-2, 1)
      else:
        N = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]
        Ofs_First_Value = Ofs_First_Value + N
 
    else:
      # for older version
      # Compute the depts of the gates
      if Dop_Model == 4000:
        for i in range(0, Nb_Gates):
          X[i] = (Indice_First_Gate + (i-1)*(Skip_Gate+1))/(2*Freq_AD)
          X[i] = (X[i] - (Retard_Flt_IQ_ns/2000000.0)) * Sound_Speed
      else:
        for i in range(0, Nb_Gates):
          X[i] = No_Gate_To_Depth_mm(i)
          
      Ofs_First_Value = Ofs_Data

    
# ----------------------------------------------------------------------

#      Select the first profile corresponding to the selected block.
#      Each profile contains additionnal byte. One of them is the block number
#
    Ofs_Start = Ofs_First_Value
    Ofs_end = Ofs_Start

    fid.seek(Ofs_Start, 0)           # point to the beginning of the data profiles
    Block_read = Get_No_Block(fid)
    if Block_read < 0:
      return 1,1,-5
    while Block_read != Block:
      N = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]
      fid.seek(N-2, 1)
      Block_read = Get_No_Block(fid)
      if Block_read < 0:
        return 1,1,-5

#   Select the first profile corresponding to the selected channel
    Channel_read = Get_No_Channel(fid)
    while Channel_read != Channel:
      N = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]
      fid.seek(N-2, 1)
      Channel_read = Get_No_Channel(fid)
      if Channel_read < 0:
        return 1,1,-4
    
#   Extract values ....
    Np = 0
    Y = []
    Info = []
    while Channel_read == Channel:
      Np = Np + 1
      Nb_Bytes_In_Profile = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]
      Gr = 1    
      while Gr < Graph:
        Gr = Gr + 1;
        Nb_Bytes_In_Graph = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]
        if Nb_Bytes_In_Graph == 0:
          return 1,1,-3
        fid.seek(Nb_Bytes_In_Graph+1, 1); # point the next graph

      Nb_Bytes_In_Graph = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]
      Data_Type = np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0]
      
      if Data_Type == Data_Type_IQ:
        Nb_Gates = Int32(Nb_Bytes_In_Graph/4)
        for i in range(0, 2*Nb_gates):
          Value = np.fromfile(fid, dtype=np.int16, count=1, sep="")[0]
          Y.append(Value)
          Value = np.fromfile(fid, dtype=np.int16, count=1, sep="")[0]
          Y.append(Value)
        # Read the additionnal bytes
        Nb_Bytes_In_Graph = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]
        Info.append(np.fromfile(fid, dtype=np.uint32, count=1, sep="")[0])  # Time stamp in ms
        Info.append(np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0])  # Block number
        Info.append(np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0])   # Mark byte
        Info.append(np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0])   # Free
        Info.append(np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0])   # Free
        Info.append(np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0])   # Number of channel
        Np = Acqui_IQ_Nprf
        break
    
      else:
          
        if Data_Type == Data_Type_Velocity:
          Nb_Gates = Nb_Bytes_In_Graph
          for i in range(0, Nb_Gates):
            Value = np.fromfile(fid, dtype=np.int8, count=1, sep="")[0]
            Value = Offset_Correction(Value,Velocity_Offset) * Coded_Value_To_Velocity
            Y.append(Value)
            
        if Data_Type == Data_Type_Echo:
          Nb_Gates = Nb_Bytes_In_Graph
          m = Module_Scale / 256
          for i in range(0, Nb_Bytes_In_Graph):
            Value = m * np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0]
            Y.append(Value)
            
        if Data_Type == Data_Type_Energy:
          Nb_Gates = Nb_Bytes_In_Graph
          m = Module_Scale / 256
          for i in range(0, Nb_Bytes_In_Graph):
            Value = m * np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0]
            Y.append(Value)
            
        if Data_Type == Data_Type_Frequency_TR1:
          Nb_Gates = Nb_Bytes_In_Graph
          for i in range(0, Nb_Bytes_In_Graph):
            Value = np.fromfile(fid, dtype=np.int8, count=1, sep="")[0]
            Value = Value * Coded_Value_To_Frequency_For_Tr1
            Y.append(Value)
            
        if Data_Type == Data_Type_Frequency_TR2:
          Nb_Gates = Nb_Bytes_In_Graph
          for i in range(0, Nb_Bytes_In_Graph):
            Value = np.fromfile(fid, dtype=np.int8, count=1, sep="")[0]
            Value = Value * Coded_Value_To_Frequency_For_Tr2
            Y.append(Value)
            
        if Data_Type == Data_Type_Frequency_TR3:
          Nb_Gates = Nb_Bytes_In_Graph
          for i in range(0, Nb_Bytes_In_Graph):
            Value = np.fromfile(fid, dtype=np.int8, count=1, sep="")[0]
            Value = Value * Coded_Value_To_Frequency_For_Tr3
            Y.append(Value)
            
        if Data_Type == Data_Type_Echo_TR1:
          Nb_Gates = Nb_Bytes_In_Graph
          m = MD_Module_Scale_Tr1 / 256
          for i in range(0, Nb_Bytes_In_Graph):
            Value = m * np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0]
            Y.append(Value)
            
        if Data_Type == Data_Type_Echo_TR2:
          Nb_Gates = Nb_Bytes_In_Graph
          m = MD_Module_Scale_Tr2 / 256
          for i in range(0, Nb_Bytes_In_Graph):
            Value = m * np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0]
            Y.append(Value)
            
        if Data_Type == Data_Type_Echo_TR3:
          Nb_Gates = Nb_Bytes_In_Graph
          m = MD_Module_Scale_Tr3 / 256
          for i in range(0, Nb_Bytes_In_Graph):
            Value = m * np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0]
            Y.append(Value)
            
        if Data_Type == Data_Type_Energy_TR1:
          Nb_Gates = Nb_Bytes_In_Graph
          m = MD_Module_Scale_Tr1 / 256
          for i in range(0, Nb_Bytes_In_Graph):
            Value = m * np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0]
            Y.append(Value)
            
        if Data_Type == Data_Type_Energy_TR2:
          Nb_Gates = Nb_Bytes_In_Graph
          m = MD_Module_Scale_Tr2 / 256
          for i in range(0, Nb_Bytes_In_Graph):
            Value = m * np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0]
            Y.append(Value)
            
        if Data_Type == Data_Type_Energy_TR3:
          Nb_Gates = Nb_Bytes_In_Graph
          m = MD_Module_Scale_Tr3 / 256
          for i in range(0, Nb_Bytes_In_Graph):
            Value = m * np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0]
            Y.append(Value)
            
        if Data_Type == Data_Type_Gate_FFT:
          Nb_Gates = Nb_Bytes_In_Graph
          m = (FFT_Scale_Pmax - FFT_Scale_Pmin)/255.0 + FFT_Scale_Pmin
          for i in range(0, Nb_Bytes_In_Graph):
            Value = m * np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0]
            Y.append(Value - Over_All_Gain_dB)
            
        if Data_Type == Data_Type_Velocity_mm_s_10:
          Nb_Gates = int32(Nb_Bytes_In_Graph/2)
          for i in range(0, Nb_Gates):
            Value = np.fromfile(fid, dtype=np.int16, count=1, sep="")[0]/10.0
            Y.append(Value)
            
        if Data_Type == Data_Type_Angle_Deg_10:
          Nb_Gates = int32(Nb_Bytes_In_Graph/2)
          for i in range(0, Nb_Gates):
            Value = np.fromfile(fid, dtype=np.int16, count=1, sep="")[0]/10.0
            Y.append(Value)
            
        if Data_Type == Data_Type_Frequency_Hz_10:
          Nb_Gates = int32(Nb_Bytes_In_Graph/2)
          for i in range(0, Nb_Gates):
            Value = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]/10.0
            Y.append(Value)
            
        if Data_Type == Data_Type_Velocity_mm_s:
          Nb_Gates = int32(Nb_Bytes_In_Graph/2)
          for i in range(0, Nb_Gates):
            Value = np.fromfile(fid, dtype=np.int16, count=1, sep="")[0]/10.0
            Y.append(Value)
            
        if Data_Type == Data_Type_Flow_ml_min:
          Nb_Gates = 1
          Value = np.fromfile(fid, dtype=np.uint32, count=1, sep="")[0]
          Y.append(Value)
    
        # Read the additionnal bytes
        Nb_Bytes_In_Graph = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]
        while Nb_Bytes_In_Graph > 0:
          fid.seek(Nb_Bytes_In_Graph+1, 1)    # point the next graph
          Nb_Bytes_In_Graph = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]

        Info.append(np.fromfile(fid, dtype=np.uint32, count=1, sep="")[0])  # Time stamp in ms
        Info.append(np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0])  # Block number
        Info.append(np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0])   # Mark byte
        Info.append(np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0])   # Free
        Info.append(np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0])   # Free
        Info.append(np.fromfile(fid, dtype=np.uint8, count=1, sep="")[0])   # Number of channel

      N = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]   #     Point to the next profile
      
      if (Mpx_Mode & 0x40000002) == 0x40000002:
        # Mpx simultaneous mode enabled ....
        try:
          for i in range (1, Number_Of_Mpx_Used_Channels):
            N = np.fromfile(fid, dtype=np.uint16, count=1, sep="")[0]
            fid.seek(N-2, 1)
        except:
          Y = np.reshape(Y,(Np, Nb_Gates))
          Info = np.reshape(Info,(Np, 6))
          return X,Y,Info,Np,Nb_Gates
            
      Channel_read = Get_No_Channel(fid)
      
    Y = np.reshape(Y,(Np, Nb_Gates))
    Info = np.reshape(Info,(Np, 6))
    return X,Y,Info,Np,Nb_Gates
#
